/*
 * Copyright (C) 2008 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file afh.c Paraslash's standalone audio format handler tool. */

#include <regex.h>

#include "para.h"
#include "string.h"
#include "afh.cmdline.h"
#include "fd.h"
#include "afh.h"
#include "error.h"
#include "version.h"
#include "ggo.h"

static struct afh_args_info conf;
INIT_AFH_ERRLISTS;

static int loglevel;
INIT_STDERR_LOGGING(loglevel)

static inline bool tag_needs_update(bool given, const char *tag,
		const char *arg)
{
	return given && (!tag || strcmp(tag, arg) != 0);
}

static int rewrite_tags(const char *name, int input_fd, void *map,
		size_t map_size, int audio_format_id, struct afh_info *afhi)
{
	struct taginfo *tags = &afhi->tags;
	bool modified = false;
	char *tmp_name;
	int output_fd = -1, ret;
	struct stat sb;

	if (tag_needs_update(conf.year_given, tags->year, conf.year_arg)) {
		free(tags->year);
		tags->year = para_strdup(conf.year_arg);
		modified = true;
	}
	if (tag_needs_update(conf.title_given, tags->title, conf.title_arg)) {
		free(tags->title);
		tags->title = para_strdup(conf.title_arg);
		modified = true;
	}
	if (tag_needs_update(conf.artist_given, tags->artist,
			conf.artist_arg)) {
		free(tags->artist);
		tags->artist = para_strdup(conf.artist_arg);
		modified = true;
	}
	if (tag_needs_update(conf.album_given, tags->album, conf.album_arg)) {
		free(tags->album);
		tags->album = para_strdup(conf.album_arg);
		modified = true;
	}
	if (tag_needs_update(conf.comment_given, tags->comment,
			conf.comment_arg)) {
		free(tags->comment);
		tags->comment = para_strdup(conf.comment_arg);
		modified = true;
	}
	if (!modified) {
		PARA_WARNING_LOG("no modifications necessary\n");
		return 0;
	}
	/*
	 * mkstmp() creates the temporary file with permissions 0600, but we
	 * like it to have the same permissions as the original file, so we
	 * have to get this information.
	 */
	if (fstat(input_fd, &sb) < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		PARA_ERROR_LOG("failed to fstat fd %d (%s)\n", input_fd, name);
		return ret;
	}
	tmp_name = make_message("%s.XXXXXX", name);
	ret = mkstemp(tmp_name);
	if (ret < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		PARA_ERROR_LOG("could not create temporary file\n");
		goto out;
	}
	output_fd = ret;
	if (fchmod(output_fd, sb.st_mode) < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		PARA_ERROR_LOG("failed to fchmod fd %d (%s)\n", output_fd,
			tmp_name);
		goto out;
	}
	ret = afh_rewrite_tags(audio_format_id, map, map_size, tags, output_fd,
		tmp_name);
	if (ret < 0)
		goto out;
	if (conf.backup_given) {
		char *backup_name = make_message("%s~", name);
		ret = xrename(name, backup_name);
		free(backup_name);
		if (ret < 0)
			goto out;
	}
	ret = xrename(tmp_name, name);
out:
	if (ret < 0 && output_fd >= 0)
		unlink(tmp_name); /* ignore errors */
	free(tmp_name);
	if (output_fd >= 0)
		close(output_fd);
	return ret;
}

static void print_info(int audio_format_num, struct afh_info *afhi)
{
	char *msg;

	afh_get_afhi_txt(audio_format_num, afhi, &msg);
	printf("%s", msg);
	free(msg);
}

static void print_chunk_table(struct afh_info *afhi)
{
	int i;

	if (conf.parser_friendly_given) {
		printf("chunk_table: ");
		for (i = 0; i <= afhi->chunks_total; i++)
			printf("%u ", afhi->chunk_table[i]);
		printf("\n");
		return;
	}
	for (i = 1; i <= afhi->chunks_total; i++) {
		struct timeval tv;
		long unsigned from, to;
		tv_scale(i - 1, &afhi->chunk_tv, &tv);
		from = tv2ms(&tv);
		tv_scale(i, &afhi->chunk_tv, &tv);
		to = tv2ms(&tv);
		printf("%d [%lu.%03lu - %lu.%03lu] %u - %u (%u)\n", i - 1,
			from / 1000, from % 1000, to / 1000, to % 1000,
			afhi->chunk_table[i - 1], afhi->chunk_table[i],
			afhi->chunk_table[i] - afhi->chunk_table[i - 1]);
	}
}

__noreturn static void print_help_and_die(void)
{
	struct ggo_help h = DEFINE_GGO_HELP(afh);
	int d = conf.detailed_help_given;
	unsigned flags = d? GPH_STANDARD_FLAGS_DETAILED : GPH_STANDARD_FLAGS;

	ggo_print_help(&h, flags);
	printf("supported audio formats: %s\n", AUDIO_FORMAT_HANDLERS);
	exit(EXIT_SUCCESS);
}

/**
 * The main function of para_afh.
 *
 * \param argc Usual argument count.
 * \param argv Usual argument vector.
 *
 * \return \p EXIT_FAILURE or \p EXIT_SUCCESS.
 */
int main(int argc, char **argv)
{
	int i, ret, audio_format_num, fd;
	void *audio_file_data;
	size_t audio_file_size;
	struct afh_info afhi;

	afh_cmdline_parser(argc, argv, &conf);
	loglevel = get_loglevel_by_name(conf.loglevel_arg);
	version_handle_flag("afh", conf.version_given);
	if (conf.help_given || conf.detailed_help_given)
		print_help_and_die();
	afh_init();
	ret = -E_AFH_SYNTAX;
	if (conf.inputs_num == 0)
		goto out;
	for (i = 0; i < conf.inputs_num; i++) {
		int ret2;
		ret = mmap_full_file(conf.inputs[i], O_RDONLY, &audio_file_data,
			&audio_file_size, &fd);
		if (ret < 0) {
			PARA_ERROR_LOG("failed to mmap \"%s\"\n", conf.inputs[i]);
			goto out;
		}
		ret = compute_afhi(conf.inputs[i], audio_file_data, audio_file_size,
			fd, &afhi);
		if (ret >= 0) {
			audio_format_num = ret;
			if (conf.modify_given) {
				ret = rewrite_tags(conf.inputs[i], fd, audio_file_data,
					audio_file_size, audio_format_num, &afhi);
			} else {
				printf("File %d: %s\n", i + 1, conf.inputs[i]);
				print_info(audio_format_num, &afhi);
				if (conf.chunk_table_given)
					print_chunk_table(&afhi);
				printf("\n");
			}
			clear_afhi(&afhi);
		}
		close(fd);
		ret2 = para_munmap(audio_file_data, audio_file_size);
		if (ret2 < 0 && ret >= 0)
			ret = ret2;
		if (ret < 0)
			break;
	}
out:
	if (ret < 0)
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	return ret < 0? EXIT_FAILURE : EXIT_SUCCESS;
}
