/*
 * Copyright (C) 2005 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file compress_filter.c Paraslash's dynamic audio range compressor. */

/*
 * Uses ideas of AudioCompress, (C) 2002-2004  M. Hari Nezumi <magenta@trikuare.cx>
 */

#include <regex.h>

#include "para.h"
#include "compress_filter.cmdline.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "buffer_tree.h"
#include "filter.h"
#include "string.h"
#include "error.h"

/** Data specific to the compress filter. */
struct private_compress_data {
	/** The current multiplier. */
	unsigned current_gain;
	/** Points to the configuration data for this instance of the compress filter. */
	struct compress_filter_args_info *conf;
	/** Maximal admissible gain. */
	unsigned max_gain;
	/** Number of samples already seen. */
	unsigned num_samples;
	/** Absolute value of the maximal sample in the current block. */
	int peak;
};

static void compress_close(struct filter_node *fn)
{
	free(fn->private_data);
}

static int compress_post_select(__a_unused struct sched *s, void *context)
{
	struct filter_node *fn = context;
	struct private_compress_data *pcd = fn->private_data;
	struct btr_node *btrn = fn->btrn;
	bool inplace = btr_inplace_ok(btrn);
	int ret;
	char *inbuf;
	size_t length, i;
	int16_t *ip, *op;
	unsigned gain_shift = pcd->conf->inertia_arg + pcd->conf->damp_arg,
		mask = (1 << pcd->conf->blocksize_arg) - 1;

	//inplace = false;
next_buffer:
	ret = btr_node_status(btrn, fn->min_iqs, BTR_NT_INTERNAL);
	if (ret < 0)
		goto err;
	if (ret == 0)
		return 0;
	btr_merge(btrn, fn->min_iqs);
	length = btr_next_buffer(btrn, &inbuf) & ~(size_t)1;
	if (length == 0) { /* eof and 1 byte available */
		ret = -E_COMPRESS_EOF;
		goto err;
	}
	ip = (int16_t *)inbuf;
	if (inplace)
		op = ip;
	else
		op = para_malloc(length);
	for (i = 0; i < length / 2; i++) {
		/* be careful in that heat, my dear */
		int sample = *ip++;
		bool neg = false;

		if (sample < 0) {
			sample = -sample;
			neg = true;
		}
		sample *= pcd->current_gain;
		sample >>= gain_shift;
		if (sample > 32767) { /* clip */
			sample = 32767;
			pcd->current_gain = (3 * pcd->current_gain +
				(1 << pcd->conf->inertia_arg)) / 4;
			pcd->peak = 0;
		} else if (sample > pcd->peak)
			pcd->peak = sample;
		op[i] = neg? -sample : sample;
		if (++pcd->num_samples & mask)
			continue;
//		PARA_DEBUG_LOG("gain: %u, peak: %u\n", pcd->current_gain,
//			pcd->peak);
		if (pcd->peak < pcd->conf->target_level_arg) {
			if (pcd->current_gain < pcd->max_gain)
				pcd->current_gain++;
		} else
			pcd->current_gain = PARA_MAX(pcd->current_gain - 2,
				1U << pcd->conf->inertia_arg);
		pcd->peak = 0;
	}
	if (inplace)
		btr_pushdown_one(btrn);
	else {
		btr_consume(btrn, length);
		btr_add_output((char *)op, length, btrn);
	}
	goto next_buffer;
err:
	assert(ret < 0);
	btr_remove_node(&fn->btrn);
	return ret;
}

/** TODO: Add sanity checks */
static int compress_parse_config(int argc, char **argv, void **config)
{
	struct compress_filter_args_info *conf = para_calloc(sizeof(*conf));

	compress_filter_cmdline_parser(argc, argv, conf);
	*config = conf;
	return 1;
}

static void compress_open(struct filter_node *fn)
{
	struct private_compress_data *pcd = para_calloc(
		sizeof(struct private_compress_data));
	pcd->conf = fn->conf;
	fn->private_data = pcd;
	fn->min_iqs = 2; /* 16 bit audio */
	pcd->current_gain = 1 << pcd->conf->inertia_arg;
	pcd->max_gain = 1 << (pcd->conf->inertia_arg + pcd->conf->aggressiveness_arg);
}

static void compress_free_config(void *conf)
{
	compress_filter_cmdline_parser_free(conf);
}

/**
 * The init function of the compress filter.
 *
 * \param f Pointer to the struct to initialize.
 */
void compress_filter_init(struct filter *f)
{
	struct compress_filter_args_info dummy;

	compress_filter_cmdline_parser_init(&dummy);
	f->open = compress_open;
	f->close = compress_close;
	f->pre_select = generic_filter_pre_select;
	f->post_select = compress_post_select;
	f->parse_config = compress_parse_config;
	f->free_config = compress_free_config;
	f->help = (struct ggo_help)DEFINE_GGO_HELP(compress_filter);
}
